/*
 * Copyright (c) 2024 ShinGeTsu Meter.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "hardware/adc.h"
#include "hardware/gpio.h"
#include "hardware/watchdog.h"
#include "pico/bootrom.h"
#include "pico/time.h"

#include "i2cDev.h"
#include "meter.h"

// GPIO
static const uint GPIO_PIN[] = {
    BTN1, BTN2, BTN3, BTN4, BTN5, BTN6, BTN7, BTN8,    // ボタン
    TA_RE1, TA_RE2,                                    // ロータリーエンコーダ
};

// DAC
int16_t dacValue = 0;    // D/Aコンバータの設定値

// 電圧値
double constV = 0;     // 定電圧
double dacV = 0;       // DAC電圧
double pcV = 0;        // PC電圧
double outputV = 0;    // メーターへの出力電圧

// メーターの目盛
double meterValue = 0;                // メーター目盛位置
double meterList[METER_HIST_SIZE];    // メーター目盛位置の履歴
double meterAve;                      // 履歴内の平均値

// TA値
double pcTA = 0;                        // PCのTA値
double taList[TA_HIST_SIZE];            // PCのTAの履歴
double taRange;                         // 履歴内の範囲
double dacTA = 0;                       // DACのTA値
double taCounter = 0;                   // TAカウンター
double counterList[COUNT_HIST_SIZE];    // TAカウンター履歴
size_t counterListIndex = 0;            // TAカウンター履歴のインデックス

// イベントフラグ
bool setEvt = false;       // セット実行
bool taMoveEvt = false;    // トーンアーム操作
bool updateEvt = false;    // LCD表示更新

// 状態管理
bool   autoSet = false;           // 自動セットモード
bool   blocking = false;          // 自動セットブロック
bool   counting = false;          // TAカウントモード
bool   pause = false;             // TAカウントを中断
bool   calibMode = false;         // 較正モード
bool   timeOver = false;          // デバッグ用 SAMPLING_uS 間隔で処理が終わらない場合true
view_t viewMode = VIEW_NORMAL;    // LCD表示モード
bool   powerView = false;         // 電圧表示

/**
 * @brief 電圧からTAを計算する。
 */
static inline double v2TA(double v) {
    return ((v / constV) * (TA_MAX - TA_MIN)) + TA_MIN;
}

/**
 * @brief 出力電圧から目盛り位置を計算する。
 */
static inline double v2meterValue(double v) {
    return (outputV) / (REFV / METER_SET);
}

/**
 * @brief TAアクションをカウントする。
 */
void countTA(bool memory) {
    static double baseTA = 0;
    if (!pause) {
        if (counting) {
            if (baseTA > dacTA) {
                taCounter += (baseTA - dacTA);
                if (memory) {
                    counterList[counterListIndex] = taCounter;
                    ++counterListIndex;
                    counterListIndex = counterListIndex % COUNT_HIST_SIZE;
                }
            }
        }
        baseTA = dacTA;
    }
}

/**
 * @brief TAカウントを履歴から復元する
 */
void undoCounter(size_t num) {
    counterListIndex = (counterListIndex + COUNT_HIST_SIZE + num) % COUNT_HIST_SIZE;
    taCounter = counterList[(counterListIndex + COUNT_HIST_SIZE - 1) % COUNT_HIST_SIZE];
}

/**
 * @brief ボタンイベントを番号毎に管理し実行を判定する。
 */
bool checkBtnAction(size_t eventNo, uint target, uint gpio, uint32_t event) {
    static uint64_t nextAction[] = {0, 0, 0, 0, 0, 0, 0, 0};
    bool            result = false;
    if (target == gpio) {
        uint64_t nowUs = time_us_64();
        result = ((nowUs > nextAction[eventNo]) && (event == GPIO_IRQ_EDGE_FALL));
        nextAction[eventNo] = nowUs + EVENT_BLOCK_uS;
    }
    return result;
}

/**
 * @brief 割り込みハンドラ
 */
void irqHandler(uint gpio, uint32_t events) {
    // ロータリーエンコーダー
    if (gpio == TA_RE1 || gpio == TA_RE2) {
        static const int8_t RE_DIR[] = ROTARY_DIRECTION;
        static uint8_t      taRE1;
        static uint8_t      taRE2;
        static size_t       reIndex = 0;

        if (setEvt) {
            setEvt = false;
            blocking = false;
        } else {
            uint8_t state = (events == GPIO_IRQ_EDGE_RISE) ? 1 : 0;
            taRE1 = (gpio == TA_RE1) ? state : taRE1;
            taRE2 = (gpio == TA_RE2) ? state : taRE2;

            size_t lastIndex = reIndex;
            reIndex = ((reIndex << 2) + (taRE1 << 1) + taRE2) & 15;
            if (lastIndex != reIndex) {
                int8_t delta = RE_DIR[reIndex];
                if (delta != 0) {
                    dacValue += delta;
                    taMoveEvt = true;
                }
            }
        }
    }

    // セットボタン Event0
    else if (checkBtnAction(0, BUTTON_SET, gpio, events)) {
        setEvt = true;
        pause = false;
    }

    // カウントモードボタン Event1
    else if (checkBtnAction(1, BUTTON_COUNT, gpio, events)) {
        if (!counting) {
            pause = false;
            countTA(false);    // 現在のTAをBaseTAとするため
            counting = true;
        } else {
            counting = false;
        }
        updateEvt = true;
    }

    // 自動モードボタン Event2
    else if (checkBtnAction(2, BUTTON_AUTO, gpio, events)) {
        autoSet = !autoSet;
        blocking = false;
        updateEvt = true;
    }

    // 表示モードボタン Event3
    else if (checkBtnAction(3, BUTTON_VIEW, gpio, events)) {
        viewMode = (++viewMode) % 3;
        calibMode = (viewMode == VIEW_NORMAL) ? false : true;
        setEvt = false;
        blocking = false;
        updateEvt = true;
    }

    // 電源表示ボタン
    else if (gpio == BUTTON_POWER) {
        powerView = (events == GPIO_IRQ_EDGE_FALL);
        updateEvt = true;
    }

    // クリアボタン Event4
    else if (checkBtnAction(4, BUTTON_CLEAR, gpio, events)) {
        countTA(false);    // 現在のTAをBaseTAとするため
        taCounter = 0;
        for (size_t i = 0; i < COUNT_HIST_SIZE; i++) {
            counterList[i] = 0.0;
        }
        pause = false;
        updateEvt = true;
    }

    // 戻る Event5
    else if (checkBtnAction(5, BUTTON_BACK, gpio, events)) {
        undoCounter(-1);
        updateEvt = true;
    }

    // 進む Event6
    else if (checkBtnAction(6, BUTTON_FOWARD, gpio, events)) {
        undoCounter(1);
        updateEvt = true;
    }
}

/**
 * @brief GPIO初期化
 */
void initIO() {
    // LED
    gpio_init(LED_PIN);
    gpio_set_dir(LED_PIN, GPIO_OUT);

    // USB接続チェック用
    gpio_init(VBUS_SENSE_PIN);
    gpio_set_dir(VBUS_SENSE_PIN, GPIO_IN);

    // 電源チェック用
    adc_init();
    adc_gpio_init(VSYS_ADC_PIN);
    adc_select_input(VSYS_ADC_NO);

    // GPIO
    uint size = sizeof(GPIO_PIN) / sizeof(uint);
    for (int i = 0; i < size; ++i) {
        gpio_init(GPIO_PIN[i]);
        gpio_set_dir(GPIO_PIN[i], GPIO_IN);
        gpio_pull_up(GPIO_PIN[i]);
        gpio_set_irq_enabled_with_callback(GPIO_PIN[i], GPIO_IRQ_EDGE_FALL, true, irqHandler);
        gpio_set_irq_enabled_with_callback(GPIO_PIN[i], GPIO_IRQ_EDGE_RISE, true, irqHandler);
    }
}

/**
 * @brief PCのTA履歴を保存する。履歴からTAの変動範囲を集計する。
 */
void addTaList(double ta) {
    static size_t listIndex = 0;

    taList[listIndex] = ta;
    double maxValue = 0.0;
    double minValue = 7.0;
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        if (maxValue < taList[i]) {
            maxValue = taList[i];
        }
        if (minValue > taList[i]) {
            minValue = taList[i];
        }
    }
    taRange = fabs(maxValue - minValue);

    listIndex = (listIndex + 1) % TA_HIST_SIZE;
}

/**
 * @brief メーター目盛の履歴を保存する。履歴から平均値を算出する。
 */
void addMeterList(double v) {
    static size_t listIndex = 0;

    meterList[listIndex] = v;
    double sumValue = 0.0;
    for (int i = 0; i < METER_HIST_SIZE; i++) {
        sumValue += meterList[i];
    }
    meterAve = sumValue / METER_HIST_SIZE;

    listIndex = (listIndex + 1) % METER_HIST_SIZE;
}

/**
 * @brief  履歴リストをクリアする
 */
void clearHist() {
    pcV = readPcV(calibMode);
    pcTA = v2TA(pcV);
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        taList[i] = pcTA;
    }
    taRange = 0.0;

    outputV = readOutputV(calibMode);
    meterValue = v2meterValue(outputV);
    for (int i = 0; i < METER_HIST_SIZE; i++) {
        meterList[i] = meterValue;
    }
    meterAve = meterValue;
}

/**
 * @brief LCDを更新する。
 */
void update() {
    char *labelCount;
    if (!counting) {
        labelCount = "     ";
    } else if (pause) {
        labelCount = "PAUSE";
    } else if (counting) {
        labelCount = "count";
    }

    char *labelAuto = "     ";
    if (setEvt) {
        labelAuto = " SET ";
    } else if (blocking) {
        labelAuto = "BLOCK";
    } else if (autoSet) {
        labelAuto = " auto";
    }
    // if (timeOver) { //サンプリング間隔調整用
    //     labelAuto =  " over";
    // }

    static char message[30];
    lcdClear();
    if (powerView) {    // 電源状態表示
        uint16_t vraw = adc_read();
        double   vsys = vraw * (3.3 * 3.0 / 4096.0);
        char    *label1;
        char    *label2;

        if (vsys < NOT_ENOUGH) {
            label1 = "NG   ";
            label2 = "   not enough   ";
        } else if (vsys < RUNNING_LOW) {
            label1 = "WARN ";
            label2 = "   running low  ";
        } else {
            label1 = "OK   ";
            label2 = "                ";
        }
        sprintf(message, "BATT:%s %4.2fv", label1, vsys);
        lcdPrint(1, message);
        lcdPrint(2, label2);
    } else if (viewMode == VIEW_NORMAL) {    // 通常表示
        sprintf(message, "%4.2f %5.2f %s", pcTA, pcTA - dacTA, labelCount);
        lcdPrint(1, message);
        sprintf(message, "%4.2f %s%6.2f", dacTA, labelAuto, taCounter);
        lcdPrint(2, message);
    } else if (viewMode == VIEW_TA) {    // 較正モード TA詳細
        sprintf(message, "%6.4f %+7.4f", pcTA, pcTA - dacTA);
        lcdPrint(1, message);
        sprintf(message, "%6.4f %4.1f %04d", dacTA, meterValue, dacValue);
        lcdPrint(2, message);
    } else if (viewMode == VIEW_VOLT) {    // 較正モード 電圧詳細
        sprintf(message, "%7.5f %7.5f", pcV, constV);
        lcdPrint(1, message);
        sprintf(message, "%7.5f %7.5f", dacV, outputV);
        lcdPrint(2, message);
    }
}

/**
 * @brief DACを0Vへ設定する
 */
void setDacZeroV() {
    dacValue = setDacValue(DAC_MIN_VALUE);
    sleep_ms(20);

    dacV = readDacV(false);
    while ((dacValue < DAC_MAX_VALUE) && (dacV < 0.0)) {
        dacValue = setDacValue(dacValue + 64);
        dacV = readDacV(false);
    }
    while ((dacValue > DAC_MIN_VALUE) && (dacV > 0.0)) {
        dacValue = setDacValue(dacValue - 16);
        dacV = readDacV(false);
    }
    while ((dacValue < DAC_MAX_VALUE) && (dacV < 0.0)) {
        dacValue = setDacValue(dacValue + 4);
        dacV = readDacV(false);
    }
    while ((dacValue > DAC_MIN_VALUE) && (dacV > 0.0)) {
        dacValue = setDacValue(dacValue - 1);
        dacV = readDacV(true);
    }
    dacValue = setDacValue(dacValue + 1);

    dacV = readDacV(true);
    dacTA = v2TA(dacV);
}

/**
 * @brief 針をセット位置へ移動する。
 */
void setNeedle() {
    // セット位置計算の校正用
    // dacValue = (pcV / constV) * fa + fb;
    static double fa = 0.0;
    static double fb = 0.0;
    // (vRate:pcV/ConstV,value:dacValue)
    static double  p1vRate = 0.0;
    static int16_t p1value = -1.0;
    static double  p2vRate = 1.0;
    static int16_t p2value = DAC_MAX_VALUE;

    bool optimization = false;

    // 初回のみ実行
    if (p1value < 0) {
        setDacZeroV();
        p1value = dacValue;
        fa = (p1value - p2value) / (p1vRate - p2vRate);
        fb = p1value - p1vRate * (p1value - p2value) / (p1vRate - p2vRate);
        optimization = true;
    }

    // SET位置へジャンプ
    constV = readConstV(true);
    dacValue = setDacValue((pcV / constV) * fa + fb);

    // 微調整
    dacV = readDacV(true);
    bool dir = pcV > dacV;
    while (setEvt) {
        int16_t tValue = dacValue;
        double  tDacV = dacV;

        dacValue = setDacValue(dacValue + (dir ? +1 : -1));
        dacV = readDacV(true);
        if (tValue == dacValue) {    // 設定可能範囲を超過
            setEvt = false;
            break;
        }

        if (dir != (pcV > dacV)) {                         // セット位置を通り過ぎたら終了
            if (fabs(pcV - tDacV) < fabs(pcV - dacV)) {    // 通り過ぎる前の方が近かったら戻す
                dacValue = setDacValue(tValue);
                dacV = readDacV(true);
            }
            optimization = true;
            setEvt = false;
        }
    }
    dacTA = v2TA(dacV);

    // SET位置計算式を最適化
    if (optimization) {
        // 近いほうを入れ替え
        if (abs(p1value - dacValue) < abs(p2value - dacValue)) {
            // 入れ替え後の差が50以下ならば入れ替えない
            if (abs(p2value - dacValue) > 50) {
                p1vRate = pcV / constV;
                p1value = dacValue;
            }
        } else {
            // 入れ替え後の差が50以下ならば入れ替えない
            if (abs(p1value - dacValue) > 50) {
                p2vRate = pcV / constV;
                p2value = dacValue;
            }
        }
        fa = (p1value - p2value) / (p1vRate - p2vRate);
        fb = p1value - p1vRate * (p1value - p2value) / (p1vRate - p2vRate);
    }
}

/**
 * @brief 状態を確認する。
 */
void checkState() {
    if (autoSet && !calibMode) {
        static uint64_t blockWaitUs = 0;
        static uint16_t outRangeNum = 0;
        double          nowUs = time_us_64();
        if (blocking) {
            if ((blockWaitUs < nowUs) && (taRange < TA_BLOCK_RANGE)) {
                blocking = false;
                setEvt = true;
                clearHist();
            }
        } else if (((taRange > TA_BLOCK_RANGE)) ||                               // TAの変動幅がTA_BLOCK_RANGEを超過
                   ((meterValue > METER_BLOCK) && meterAve < METER_BLOCK_TH))    // 急激にメータ表示範囲を超過
        {
            blocking = true;
            pause = true;
            dacValue = setDacValue(DAC_MIN_VALUE);
            blockWaitUs = nowUs + BLOCK_uS;
            updateEvt = true;
        }
        // 自動セットの閾値超過 → 自動セット
        else if ((meterAve <= METER_AUTO_SET_MIN) || (meterAve >= METER_AUTO_SET_MAX)) {
            outRangeNum++;
            if (AUTO_SET_WAIT_NUM <= outRangeNum) {
                setEvt = true;
            }
        } else {
            outRangeNum = 0;
        }
    }
}

/**
 * @brief メイン
 */
int main() {
    sleep_ms(100);
    initIO();
    initI2C();

    // 電源電圧を表示
    powerView = true;
    update();
    uint64_t waitUs = time_us_64() + POWER_VIEW_WAIT_uS;
    while (time_us_64() < waitUs) {    // 表示中にBTN1が押された場合DFUモードへ移行
        if (!gpio_get(BUTTON_SET) && gpio_get(VBUS_SENSE_PIN)) {
            sleep_ms(50);
            lcdClear();
            lcdPrint(1, "Device Firmware");
            lcdPrint(2, "Update");
            reset_usb_boot(0, 0);
        }
        sleep_ms(1);
    }
    powerView = false;

    viewMode = VIEW_NORMAL;
    autoSet = true;
    setEvt = true;
    taMoveEvt = false;
    updateEvt = false;
    clearHist();
    for (size_t i = 0; i < COUNT_HIST_SIZE; i++) {
        counterList[i] = 0.0;
    }
    constV = readConstV(true);

    uint64_t nextReadUs = 0;
    uint64_t netxUpdateUs = 0;
    while (true) {
        while (time_us_64() < nextReadUs) {
            if (taMoveEvt || setEvt || updateEvt) {
                break;
            }
            sleep_us(500);
        }

        if (calibMode) {
            constV = readConstV(true);
            dacV = readDacV(true);
        }

        pcV = readPcV(calibMode);
        pcTA = v2TA(pcV);
        addTaList(pcTA);
        outputV = readOutputV(calibMode);
        meterValue = v2meterValue(outputV);
        addMeterList(meterValue);
        checkState();

        if (taMoveEvt) {
            dacValue = setDacValue(dacValue);
            dacV = readDacV(true);
            dacTA = v2TA(dacV);
            countTA(false);
            update();
            taMoveEvt = false;
        } else if (setEvt) {
            update();
            gpio_put(LED_PIN, true);
            setNeedle();
            clearHist();
            countTA(true);
            gpio_put(LED_PIN, false);
            blocking = false;
            update();
        } else if (updateEvt) {
            update();
            updateEvt = false;
        } else {
            uint64_t nowUs = time_us_64();
            nextReadUs += SAMPLING_uS;
            if (nextReadUs < nowUs) {
                nextReadUs = nowUs + SAMPLING_uS;
                timeOver = true;
            } else {
                timeOver = false;
            }

            if (netxUpdateUs < nowUs) {
                update();
                netxUpdateUs += UPDATE_uS;
            }
            if (netxUpdateUs < nowUs) {
                netxUpdateUs = nowUs + UPDATE_uS;
            }
        }
    }
    return 0;
}